#-----------------------------------------------------------------------------#
# University of Milan: DAPS&CO 2021
# - Social Network Analysis   (Class 5)
# Instructor: Rob Fahey <robfahey@aoni.waseda.jp>
#-----------------------------------------------------------------------------#

Sys.setlocale("LC_ALL", 'en_GB.UTF-8')
Sys.setenv(LANG = "en_US.UTF-8")
getwd()
setwd("~/Dropbox/Milan/2021 Social Network Analysis/Class 5 - Week 1 Friday/")   
# Set the above to the folder on your computer where you've saved the files for this class.

#-----------------------------------------------------------------------------#
# 1.  A More Complex Dataset: Game of Thrones
#-----------------------------------------------------------------------------#

# Let's start by importing some different sample data. We've got a CSV file which shows all of the 
# interactions in Game of Thrones (the first book of the Song and Ice and Fire series by George R. R. 
# Martin - the interactions are based on the books, not the TV show). 

# This is a good data set to try out Community Detection algorithms on, because as anyone who has
# read these books or watched the TV show knows, there are a huge number of characters and quite a
# few different locations in the series. This should be a far more complex network than Star Wars was!

# Let's start by opening it up and seeing what kind of data is in there.

library(igraph)

got <- read.csv("got_1.csv")
head(got)


# This edgelist has a bit more information - some of which we don't need.
# - Source and Target for each edge; 
# - edge type (these are # all undirected so it doesn't really matter); 
# - weight indicating how many times the characters interact in the book;
# - "book" is which book of the series this is from, but we're only looking at Book 1 here anyway.

# Let's tidy it up a bit. We want to remove columns that aren't relevant, and ideally we should shorten
# some of those ridiculously long names. Since we know that all the characters have unique first names,
# we can just keep the first part of each name and drop the surnames and titles. This will make it much
# easier to plot a tidy graph later!

library(dplyr)
library(tidyr)

got <- got %>% 
  separate(Source, c("Source"), extra="drop") %>%
  separate(Target, c("Target"), extra="drop") %>%
  select(-Type, -book)

# We also want to rename that "weight" column, to avoid confusion with different interpretations of
# what "weight" means. Like the Star Wars data, in this dataset it refers to scenes where the
# characters appear together.

colnames(got)[3] <- "scenes"

# Okay, now we make an undirected graph from the data.

got_net <- graph_from_data_frame(d = got, directed = FALSE)

plot(got_net)

# Well, that's a mess, but that's good for our purposes - it means we're dealing with a more complex
# dataset, so there's value to doing more advanced analysis of this network.

# To get a visual idea of how complex this dataset is...
plot(got_net, layout = layout_in_circle)

# Perhaps more usefully, we can explore that complexity in statistics.
# Network Size:
vcount(got_net)

# Edge Count:
ecount(got_net)

# Edge Density:
edge_density(got_net)

# Network Diameter:
diameter(got_net)

# Mean Distance:
mean_distance(got_net)


# Okay, what have we learned from this set of statistics?

# We know that our network is large compared to anything we've analysed before (182 nodes - similar in size to the
# parliament of many smaller countries or regional assemblies, in fact), but it's also relatively sparse. The
# edge density is only 0.04, which means there are a lot of characters who do not interact at all.

# This is also confirmed by the diameter and mean distance measurements. The mean distance is only around half the
# diameter, which means that most nodes are probably inside local clusters (i.e. groups of other nodes to which
# they have a short distance), but that there are multiple clusters in the network overall (so the total possible
# diameter is high).

# Even if we knew nothing about the actual data contained in this graph, these statistics alone would hint strongly
# at the existence of clusters in the network, meaning that a community detection approach is valid.



# Before we do anything, though, we should take one quick step to prepare our data. Community 
# detection generally doesn't work well on graphs which have either loops (i.e. a node that links 
# back to itself), or multiple edges (i.e. more than one edge connecting the same pair of nodes).
# These things can happen in your data for various reasons; often they're just an accident or a typo.
# igraph has a function to detect and delete these problems easily:

got_net <- simplify(got_net, remove.multiple = TRUE, remove.loops = TRUE,
                    edge.attr.comb = "sum")

# Note the option "edge.attr.comb" - this specifies what we should do with the edge attributes (such as
# "scenes") when we encounter multiple edges. Here, we're telling igraph to add them together, creating 
# a new edge with the scenes from both original edges.


# Let's start from a reasonably legible plot of the graph - we'll save this layout and use it to compare the
# results of the various methods we try in this class.


set.seed(555)
l <- layout_with_fr(got_net, weights = E(got_net)$scenes)
l <- norm_coords(l, ymin=-1, ymax=1, xmin=-1, xmax=1)
plot(got_net, vertex.size = 5, vertex.label.cex = .4, rescale = FALSE, layout = l*1.2)



#-----------------------------------------------------------------------------#
# 2.  Community Detection Basics
#-----------------------------------------------------------------------------#

# Community detection, sometimes called "modularity", is a method that aims to try to find densely
# interconnected communities within the wider structure of the graph.

# Like the layout algorithms we looked at yesterday, there are a lot of different community detection
# algorithms, but they all go back to the same basic mathematical concept: randomness.

# Essentially, we say that a group of nodes in a graph could be a community if the number of edges
# within that group is higher than the number of edges we would expect to see if connections were
# spread randomly around in a network of that size and density.

# (This kind of probabilistic approach should be familiar to you from other statistical methods you
# have studied; most statistical proofs are fundamentally about showing that it's very unlikely that
# the effect you are observing could have happened at random. Community detection within a network
# graph is no different.)



# Let's look at a few of the most popular algorithms.


# EDGE BETWEENNESS
# ----------------

# We've discussed Edge Betweenness in a previous class. This approach attempts to find communities
# in the network by removing the edges with the highest betweenness - i.e. those with the most
# geodesics running over them - in an attempt to see what communities are created by doing so.

groups <- cluster_edge_betweenness(got_net)
show(groups)

# Running Edge Betweenness with no extra parameters on this data gives us 11 groups in total.
# Lets take a look at them on the graph. This plot command is exactly the same as the one
# we were using previously, the only difference being that we also pass the "groups" object
# as the first parameter.

plot(groups, got_net, vertex.size = 5, vertex.label.cex = .4, rescale = FALSE, layout = l*1.2)


# Edge Betweenness can, of course, also take into account edge weights. However, if we just
# directly pass it the "scenes" parameter, that won't work very well, because it's expecting the
# weights to represent distance / friction, not the strength / importance of the connection.

# There are lots of different ways you could choose to convert edge strengths into edge distances.
# The most simple is just (1 / weight), i.e. inverting the value.
# However in this case the maximum weight is 291 and the minimum is 1, which is a huge difference,
# so I'm choosing to flatten it slightly by first dividing the weights by the max weight.

E(got_net)$distances <- 1 / (E(got_net)$scenes / max(E(got_net)$scenes))


# Let's try re-running the clustering algorithm with these weights...
groups <- cluster_edge_betweenness(got_net, weights = E(got_net)$distances)

# We get some warnings, which is actually... Not surprising, because, even igraph gets internally
# confused about some of its functions treating weights as distances, and some treating weights
# as importances. Oops. 

show(groups)
plot(groups, got_net, vertex.size = 5, vertex.label.cex = .4, rescale = FALSE, layout = l*1.2)


# WALKTRAP
# --------

# The Walktrap algorithm takes a different approach. It runs a very large number of random walks
# (i.e. picking its next step randomly), on the assumption that if a node is a member of a 
# community, a disproportionate number of those random walks should remain within the densely
# connected community, compared to the small number that will actually leave the community.

groups <- cluster_walktrap(got_net)
show(groups)
plot(groups, got_net, vertex.size = 5, vertex.label.cex = .4, rescale = FALSE, layout = l*1.2)

# The main parameter we can change for the walktrap algorithm is the "steps" parameter. By
# default, it creates paths just four steps long - which in a large and complex graph might result
# in over- or under-specifying the communities.

groups <- cluster_walktrap(got_net, steps = 8)
show(groups)
plot(groups, got_net, vertex.size = 5, vertex.label.cex = .4, rescale = FALSE, layout = l*1.2)


# FAST-GREEDY
# -----------

# The unusually named "Fast and Greedy" algorithm basically works by trying to continually
# optimise a modularity score for the network at each step of iteration.

groups <- cluster_fast_greedy(got_net)
show(groups)
plot(groups, got_net, vertex.size = 5, vertex.label.cex = .4, rescale = FALSE, layout = l*1.2)

# Note that Fast-Greedy seems to have found far fewer groups than the others we looked at so far.
# It's got less of a tendency to generate "tiny" groups of just one or two nodes.

# With weights:
groups <- cluster_fast_greedy(got_net, weights = E(got_net)$scenes)
show(groups)
plot(groups, got_net, vertex.size = 5, vertex.label.cex = .4, rescale = FALSE, layout = l*1.2)


# LABEL PROPAGATION
# -----------------

# Label Propagation is a quite physics-based approach to community detection. You randomly 
# seed the network with labels, and then over the course of hundreds of iterations, you
# change the label of each node to the dominant label of all the nodes surrounding it.
# The effect is a bit like pouring water into an uneven object; each label finds its 
# natural "level" within the graph.

groups <- cluster_label_prop(got_net)
show(groups)
plot(groups, got_net, vertex.size = 5, vertex.label.cex = .4, rescale = FALSE, layout = l*1.2)

# In this case, Label Propagation has arguably identified too few clusters to be useful. 
# Even though our data isn't all that dense, it has created one huge cluster and only a few
# small peripheral clusters, which probably isn't what we'd want in most situations.

groups <- cluster_label_prop(got_net, weights = E(got_net)$scenes)
show(groups)
plot(groups, got_net, vertex.size = 5, vertex.label.cex = .4, rescale = FALSE, layout = l*1.2)


# INFOMAP
# -------

# Last but by no means least, Infomap is a complex and powerful algorithm which applies an 
# information theory model to the problem. It's similar to Walktrap in that it's based on 
# performing random walks through the data, but different in that it tries to identify areas
# where a lot of random walks get "trapped" and collapse them down, gradually simplifying the
# graph into a set of hierarchical clusters.

groups <- cluster_infomap(got_net)
show(groups)
plot(groups, got_net, vertex.size = 5, vertex.label.cex = .4, rescale = FALSE, layout = l*1.2)

# Uniquely, Infomap allows us to set weights for both the edges and the nodes in the network.

# Just edges...
groups <- cluster_infomap(got_net, e.weights = E(got_net)$scenes)
show(groups)
plot(groups, got_net, vertex.size = 5, vertex.label.cex = .4, rescale = FALSE, layout = l*1.2)

# Now let's try also setting the strength() of the nodes as their weight.
# (The weight of a node is the a priori likelihood that a random walk will jump to that node in its next step.)
groups <- cluster_infomap(got_net, e.weights = E(got_net)$scenes, v.weights = strength(got_net, weights = E(got_net)$scenes))
show(groups)
plot(groups, got_net, vertex.size = 5, vertex.label.cex = .4, rescale = FALSE, layout = l*1.2)


#-----------------------------------------------------------------------------#
# 3.  What can we do with Communities?
#-----------------------------------------------------------------------------#


# Let's pick one of the community groupings which seems to do a decent job with this data, and explore
# what we can actually do with this new information.

# Just looking at the plots for each one, I think the weighted "fast_greedy" algorithm seems to do a decent
# job of dividing the data into a reasonable number of sensible looking communities, so we'll use that:

groups <- cluster_fast_greedy(got_net, weights = E(got_net)$scenes)
show(groups)
plot(groups, got_net, vertex.size = 5, vertex.label.cex = .4, rescale = FALSE, layout = l*1.2)

# The first step is to create a new parameter for the nodes in our network, indicating which
# community they seem to be a member of.

V(got_net)$community <- groups$membership

# We can also use this value to assign a colour to each group automatically.
# Let's use the same palette from RColorBrewer we were using yesterday.

library(RColorBrewer)

V(got_net)$color <- brewer.pal(n=8, name="Set2")[V(got_net)$community]


# Now we could, for example, use this new label to look at each community individually.
# For example, here's Community 1 in the graph...
plot(induced_subgraph(got_net, V(got_net)$community == 1))

# And here's Community 2...
plot(induced_subgraph(got_net, V(got_net)$community == 2))


# What might be more useful is to get some summary statistics about all of the communities, so
# we can start to understand the "shape" of our newly partitioned data.

subcom <- data.frame()
for (i in unique(V(got_net)$community)) {
  this_subgraph <- induced_subgraph(got_net, V(got_net)$community == i)
  size <- vcount(this_subgraph)
  centrality <- eigen_centrality(this_subgraph)$vector
  subcom <- bind_rows(subcom, data.frame(community = i, size = size, central = names(which(centrality == max(centrality)))))
}
show(subcom)

# If you've read / watched Game of Thrones, this might be the first point where you've looked at 
# this data and gone "okay, this is starting to make sense". Each one of these characters (with the
# exception of the last two) is a major, central character in Game of Thrones, and moreover, each of
# them moves through a quite different social group in the book - in some cases, even being physically
# located in a separate other location from other characters for much of the story.

# We also have two very small communities (with only three people in each one). Much like isolates or 
# small connected components in a network, very small communities can also often be ignored or entirely
# removed in analysis - though of course, this depends on exactly what kind of analysis you're trying
# to conduct in the first place.


#-----------------------------------------------------------------------------#
# 4.  Plotting Tidier Graphs with Communities
#-----------------------------------------------------------------------------#


# Let's try to tidy up our graph using this new information we got about the communities.

# First, we'll make a list of the top three characters in each large community, and set those to the
# $label value in the graph - so we'll only be plotting the names of the top three characters in each
# community.

labels_to_plot <- c()
for (i in unique(V(got_net)$community)) {
  this_subgraph <- induced_subgraph(got_net, V(got_net)$community == i)
  if (vcount(this_subgraph) >= 10) {
    centrality <- eigen_centrality(this_subgraph)$vector
    labels_to_plot <- append(labels_to_plot, names(head(sort(centrality, decreasing = TRUE), 3)))
  }
}
show(labels_to_plot)
V(got_net)$label <- ifelse(V(got_net)$name %in% labels_to_plot,
                           V(got_net)$name,
                           NA)

plot(got_net, vertex.size = 5, vertex.label.cex = .75, rescale = FALSE, layout = l*1.2)

# Now let's get rid of those small (and probably unimportant) communities:
#     (and let's also apply some of the visualisation tricks we used yesterday)

got_net_tidy <- delete_vertices(got_net, V(got_net)[community %in% c(6, 8)])

# Drop edges that have the minimum weight (in this case, 3):
got_net_tidy <- delete_edges(got_net_tidy, E(got_net_tidy)[scenes <= 3])

# And finally remove any isolates this has created (i.e. nodes that no longer
# have any connections at all):
got_net_tidy <- delete_vertices(got_net_tidy, which(degree(got_net_tidy) == 0))


plot(got_net_tidy, 
     vertex.size = 8, 
     layout = layout_with_kk,
     edge.curved = .2,
     edge.width = E(got_net_tidy)$scenes*0.2,
     vertex.label.cex = .75,
     vertex.label.color = "black",
     vertex.frame.color = "lightgrey")


#-----------------------------------------------------------------------------#
# Today's Assignment
#-----------------------------------------------------------------------------#

# Today's assignment is to carry out a community analysis process on another data file - this time
# one taken from the real world, namely a relationship graph of all of the people with business,
# political or family relationships to former U.S. president Donald Trump.

# This data was originally compiled by reporters for Buzzfeed News who set out to map the network
# of relationships that made up what they called "TrumpWorld".

# The file is "trumpworld.csv" - try to replicate some of our analysis from the class and discover
# what kinds of communities seem to exist inside "TrumpWorld".

# Your tasks are...

# - Turn the Trumpworld data into a network and find an effective community detection algorithm for it.

# - Try to figure out what each community in your data represents. You can explore it by seeing which
#   individuals appear prominently, and looking at the information about those individuals contained
#   in the "Relationship" attribute.

# - Finally, try to make a good quality visualisation of Trumpworld which shows all of the communities
#   and how they are interconnected.

# This assignment is worth 15% of your final grade and is due by Monday morning.


